'''Utility functions, many for reading the CHIANTI database files.

Copyright 2009, 2010 Kenneth P. Dere

This software is distributed under the terms of the GNU General Public License
that is found in the LICENSE file


'''
import os, fnmatch
from types import *
#from ConfigParser import *
import pickle
from datetime import date
import numpy as np
from scipy import interpolate
from .FortranFormat import *
from . import _chianti_constants as const
#
#
def between(array,limits):
    '''returns an index array of elements of array where the values lie
    between the limits given as a 2 element list or tuple'''
    array=np.asarray(array)
    nlines=len(array)
    hi=np.where(array >= limits[0],range(1,nlines+1),0)
    lo=np.where(array <= limits[1],range(1,nlines+1),0)
    hilo=hi&lo
    out=[a -1  for a in hilo if a > 0]
    return out
    # -------------------------------------------------------------------------------------
    #
def ipRead(verbose=False):
    """
    reads the ionization potential file, returns ip array in eV
    """
    topdir=os.environ["XUVTOP"]
    ipname=os.path.join(topdir, 'ip','chianti.ip')
    ipfile=open(ipname)
    data=ipfile.readlines()
    ipfile.close()
    nip=0
    ndata=2
    maxz=0
    while ndata > 1:
        s1=data[nip]
        s2=s1.split()
        ndata=len(s2)
        nip=nip+1
        if int(s2[0]) > maxz:
            maxz=int(s2[0])
    if verbose:
        print((' maxz = %5i'%(maxz)))
    nip=nip-1
    ip=np.zeros((maxz, maxz), 'Float64')
    for aline in data[0:nip]:
        s2=aline.split()
        iz=int(s2[0])
        ion=int(s2[1])
        ip[iz-1, ion-1]=float(s2[2])
    return ip*const.invCm2Ev
    #
    # -------------------------------------------------------------------------------------
    #
def masterListInfo(force=0):
    """ returns information about ions in masterlist
    the reason for this file is to speed up multi-ion spectral calculations
    the information is stored in a pickled file 'masterlist_ions.pkl'
    if the file is not found, one will be created and the following information
    returned for each ion
    wmin, wmax :  the minimum and maximum wavelengths in the wgfa file
    tmin, tmax :  the minimum and maximum temperatures for which the ionization balance is nonzero"""
    dir=os.environ["XUVTOP"]
    infoPath = os.path.join(dir, 'masterlist')
    infoName=os.path.join(dir,'masterlist','masterlist_ions.pkl')
    masterName=os.path.join(dir,'masterlist','masterlist.ions')
    #
    makeNew = force == 1 or not os.path.isfile(infoName)
#    if os.path.isfile(infoName):
    if not makeNew:
#       print ' file exists - ',  infoName
        pfile = open(infoName, 'r')
        masterListInfo = pickle.load(pfile)
        pfile.close
    elif os.access(infoPath, os.W_OK):
        # the file does not exist but we have write access and will create it
        defaults = defaultsRead()
        print((' defaults = %s'%(str(defaults))))
        ioneqName = defaults['ioneqfile']
        ioneq = ioneqRead(ioneqname = ioneqName)
        masterList = masterListRead()
        masterListInfo = {}
        haveZ = [0]*31
        haveStage = np.zeros((31, 31), 'Int32')
        haveDielectronic = np.zeros((31, 31), 'Int32')
        for one in masterList:
            ionInfo = convertName(one)
            z = ionInfo['Z']
            stage = ionInfo['Ion']
            haveZ[z] = 1
            dielectronic = ionInfo['Dielectronic']
            if dielectronic:
                haveDielectronic[z, stage] = 1
            else:
                haveStage[z, stage] = 1
            thisIoneq = ioneq['ioneqAll'][z- 1, stage - 1 + dielectronic]
            good = thisIoneq > 0.
            goodTemp = ioneq['ioneqTemperature'][good]
            tmin = goodTemp.min()
            tmax = goodTemp.max()
            vgood = thisIoneq == thisIoneq.max()
            vgoodTemp = ioneq['ioneqTemperature'][vgood][0]
            wgfa = wgfaRead(one)
            nZeros = wgfa['wvl'].count(0.)
            # two-photon transitions are denoted by a wavelength of zero (0.)
            while nZeros > 0:
                wgfa['wvl'].remove(0.)
                nZeros = wgfa['wvl'].count(0.)
            # unobserved lines are denoted with a negative wavelength
            wvl = np.abs(np.asarray(wgfa['wvl'], 'float64'))
            wmin = wvl.min()
            wmax = wvl.max()
            masterListInfo[one] = {'wmin':wmin, 'wmax':wmax, 'tmin':tmin, 'tmax':tmax, 'tIoneqMax':vgoodTemp}
        masterListInfo['haveZ'] = haveZ
        masterListInfo['haveStage'] = haveStage
        masterListInfo['haveDielectronic'] = haveDielectronic
        #  now do the bare ions from H thru Zn
        #  these are only involved in the continuum
        for iz in range(1, 31):
            ions = zion2name(iz, iz+1)
            thisIoneq = ioneq['ioneqAll'][iz-1, iz]
            good = thisIoneq > 0.
            goodTemp = ioneq['ioneqTemperature'][good]
            tmin = goodTemp.min()
            tmax = goodTemp.max()
            wmin=0.
            wmax = 1.e+30
            masterListInfo[ions] = {'wmin':wmin, 'wmax':wmax, 'tmin':tmin, 'tmax':tmax}
        pfile = open(infoName, 'w')
        pickle.dump(masterListInfo, pfile)
        pfile.close
    else:
        # the file does not exist and we do NOT have write access to creat it
        # will just make an inefficient, useless version
        masterListInfo = {}
        for one in masterList:
            ionInfo = convertName(one)
            z = ionInfo['Z']
            stage = ionInfo['Ion']
            dielectronic = ionInfo['Dielectronic']
            wmin=0.
            wmax = 1.e+30
            masterListInfo[one] = {'wmin':wmin, 'wmax':wmax, 'tmin':1.e+4, 'tmax':1.e+9}
        #  now do the bare ions from H thru Zn
        #  these are only involved in the continuum
        for iz in range(1, 31):
            ions = zion2name(iz, iz+1)
            wmin=0.
            wmax = 1.e+30
            masterListInfo[ions] = {'wmin':wmin, 'wmax':wmax, 'tmin':1.e+4, 'tmax':1.e+9}
        pfile = open(infoName, 'w')
        pickle.dump(masterListInfo, pfile)
        pfile.close
        masterListInfo = {'noInfo':'none'}
    return masterListInfo
    #
    # -------------------------------------------------------------------------------------
    #
def masterListRead():
    """ read a chianti masterlist file and return a list of files"""
    dir=os.environ["XUVTOP"]
    fname=os.path.join(dir,'masterlist','masterlist.ions')
    input=open(fname,'r')
    s1=input.readlines()
    dum=input.close()
    masterlist=[]
    for i in range(0,len(s1)):
        s1a=s1[i][:-1]
        s2=s1a.split(';')
        masterlist.append(s2[0].strip())
    return masterlist
    #
    # -------------------------------------------------------------------------------------
    #
def photoxRead(ions):
    """read chianti photoionization .photox files and return
        {"energy", "cross"} where energy is in Rydbergs and the
        cross section is in cm^2  """
    #
    zion=convertName(ions)
    if zion['Z'] < zion['Ion']:
        print((' this is a bare nucleus that has no ionization rate'))
        return
    #
    fname=ion2filename(ions)
    paramname=fname+'.photox'
    input=open(paramname,'r')
    lines = input.readlines()
    input.close
    # get number of energies
#    neng = int(lines[0][0:6])
    dataEnd = 0
    lvl1 = []
    lvl2 = []
    energy = []
    cross = []
    icounter = 0
    while not dataEnd:
        lvl11 = int(lines[icounter][:8])
        lvl21 = int(lines[icounter][8:15])
        ener = lines[icounter][15:].split()
        energy1 = np.asarray(ener, 'float64')
        #
        icounter += 1
        irsl = int(lines[icounter][:8])
        ind0 = int(lines[icounter][8:15])
        if irsl != lvl11 or ind0 != lvl21:
            # this only happens if the file was written incorrectly
            print((' lvl1, lvl2 = %7i %7i'%(lvl11, lvl21)))
            print((' irsl, indo = %7i %7i'%(irsl,  ind0)))
            return
        crs = lines[icounter][15:].split()
        cross1 = np.asarray(crs, 'float64')
        lvl1.append(lvl11)
        lvl2.append(lvl21)
        energy.append(energy1)
        cross.append(cross1)
        icounter += 1
        dataEnd = lines[icounter].count('-1')
    ref = lines[icounter+1:-1]
    cross = np.asarray(cross, 'float64')
    energy = np.asarray(energy, 'float64')
    return {'lvl1':lvl1, 'lvl2':lvl2,'energy':energy, 'cross':cross,  'ref':ref}
    #
    # -------------------------------------------------------------------------------------
    #
def ionrecdatRead(filename):
    """ read chianti ionxdat, ionizdat, recombdat files and return
    {"ev":ev,"cross":cross,"crosserr":crosserr,"ref":ref}  not tested """
    #
    input=open(filename,'r')
    ionrec=input.readlines()
    dum=input.close()
    #
    # first get the number of data lines
    ndata=2
    iline=0
    while ndata > 1:
        s2=ionrec[iline].split()
        ndata=len(s2)
        iline=iline+1
    nline=iline-1
    #
    x=np.zeros(nline,'Float64')
    y=np.zeros(nline,'Float64')
    yerr=np.zeros(nline,'Float64')
#
    for iline in range(0,nline):
        ndata=len
        s2=ionrec[iline].split()
        ndata=len(s2)
        if ndata == 2:
            x[iline]=float(s2[0])
            y[iline]=float(s2[1])
            yerr[iline]=float(0.)
        else:
            x[iline]=float(s2[0])
            y[iline]=float(s2[1])
            yerr[iline]=float(s2[2])
    #
    ref=[]
    for iline in range(nline+1,len(ionrec)-1):
        s1a=ionrec[iline][:-1]
        ref.append(s1a.strip())


    ionrecdat={"x":x,"y":y,"yerr":yerr,"ref":ref}
    return ionrecdat
    #
    # --------------------------------------------------
    #
def z2element(z):
    """ convert Z to element string """
    if z-1 < len(const.El):
        thisel=const.El[z-1]
    else:
        thisel=''
    return thisel
    #
    # -------------------------------------------------------------------------------------
    #
def zion2name(z,ion, dielectronic=False):
    """ convert Z, ion to generic name  26, 13 -> fe_13 """
    if (z-1 < len(const.El)) and (ion <= z+1):
        thisone=const.El[z-1]+'_'+str(ion)
        if dielectronic:
            thisone+='d'
    else:
        thisone=''
    return thisone
    #
    # -------------------------------------------------------------------------------------
    #
def zion2filename(z,ion, dielectronic=False):
    """ convert Z to generic file name string """
    dir=os.environ["XUVTOP"]
    if (z-1 < len(const.El)) and (ion <= z+1):
        thisel=const.El[z-1]
    else:
        thisel=''
    if z-1 < len(const.El):
        thisone=const.El[z-1]+'_'+str(ion)
        if dielectronic:
            thisone+='d'
    else:
        thisone=''
    if thisel != '' :
        fname=os.path.join(dir,thisel,thisone,thisone)
    return fname
    #
    # -------------------------------------------------------------------------------------
    #
def zion2localFilename(z,ion, dielectronic=False):
    """ convert Z to generic file name string with current directory at top"""
    dir='.'
    if (z-1 < len(const.El)) and (ion <= z+1):
        thisel=const.El[z-1]
    else:
        thisel=''
    if z-1 < len(const.El):
        thisone=const.El[z-1]+'_'+str(ion)
        if dielectronic:
            thisone+='d'
    else:
        thisone=''
    if thisel != '' :
        fname=os.path.join(dir,thisel,thisone,thisone)
    return fname
    #
    # -------------------------------------------------------------------------------------
    #
def zion2spectroscopic(z,ion, dielectronic=False):
    """ convert Z and ion to spectroscopic notation string """
    if (z-1 < len(const.El)) and (ion <= z+1):
        spect=const.El[z-1].capitalize()+' '+const.Ionstage[ion-1]
        if dielectronic:
            spect+=' d'
    else:  spect = ''
    return spect
    #
    # -------------------------------------------------------------------------------------
    #
def convertName(name):
    """ convert ion name string to Z and Ion """
    s2=name.split('_')
    els=s2[0].strip()
    i1=const.El.index(els)+1
    ions=s2[1].strip()
    d=ions.find('d')
    if d >0 :
        dielectronic=True
        ions=ions.replace('d','')
    else: dielectronic=False
    return {'Z':int(i1),'Ion':int(ions),'Dielectronic':dielectronic, 'Element':els}
    #
    # -------------------------------------------------------------------------------------
    #
def defaultsRead(verbose=0):
    #
    #possibleDefaults = {'wavelength':['angstrom', 'kev', 'nm']}
    #symbolDefaults = {'wavelength':['A', 'keV', 'nm']}
    initDefaults={'abundfile': 'sun_photospheric_1998_grevesse','ioneqfile': 'chianti', 'wavelength': 'angstrom', 'flux': 'energy','gui':False}
    rcfile=os.path.join(os.environ['HOME'],'.chianti/chiantirc')
    if os.path.isfile(rcfile):
        print((' reading chiantirc file'))
        config = RawConfigParser(initDefaults)
        config.read(rcfile)
        defaults = {}
        for anitem in config.items('chianti'):
            defaults[anitem[0]] = anitem[1]
        if defaults['gui'].lower() in ('t', 'y', 'yes', 'on', 'true', '1', 1, True):
            defaults['gui'] = True
        elif defaults['gui'].lower() in ('f', 'n', 'no', 'off', 'false', '0', 0, False):
            defaults['gui'] = False
    else:
        defaults = initDefaults
        if verbose:
            print((' chiantirc file (/HOME/.chianti/chiantirc) does not exist'))
            print((' using the following defaults'))
            for akey in list(defaults.keys()):
                print((' %s = %s'%(akey, defaults[akey])))
    return defaults
    #
    # -------------------------------------------------------------------------------------
    #
def ion2filename(ions):
    """ convert ion string to generic file name string """
    dir=os.environ["XUVTOP"]
    zion=convertName(ions)
    el=z2element(zion['Z'])
    fname=os.path.join(dir,el,ions,ions)
    return fname
    #
    # -------------------------------------------------------------------------------------
    #
def el2z(els):
    """ from an the name of the element (1-2 letter) return Z"""
    z=const.El.index(els.lower())+1
    return z
    #
    # -------------------------------------------------------------------------------------
    #
def abundanceRead(abundancename=''):
    """ read an abundanc file and returns the abundance values relative to hydrogen"""
    pass

    #
    # -------------------------------------------------------------------------------------
    #
def qrp(z,u):
    ''' qrp(Z,u)  u = E/IP
    calculate Qr-prime (equ. 2.12) of Fontes, Sampson and Zhang 1999'''
    #
    aa=1.13  # aa stands for A in equ 2.12
    #
    if z >= 16 :
        # use Fontes Z=20, N=1 parameters
        dd=3.70590
        c=-0.28394
        d=1.95270
        cc=0.20594
    else:
    # use Fontes Z=10, N=2 parameters
        dd=3.82652
        c=-0.80414
        d=2.32431
        cc=0.14424
    #
    if z > 20:
        cc+=((z-20.)/50.5)**1.11
    #
    bu=u <= 1.
    q=np.ma.array(u, 'Float64', mask=bu, fill_value=0.)
    #
    #
    q=(aa*np.ma.log(u) + dd*(1.-1./u)**2 + cc*u*(1.-1./u)**4 + (c/u+d/u**2)*(1.-1/u))/u
    #
    q.set_fill_value(0.)  # I don't know why this is necessary
    return q  #  .set_fill_value(0.)
    #
    # -------------------------------------------------------------------------------------
    #
def elvlcRead(ions, filename = None, verbose=0,  useTh=1):
    """
    read a chianti energy level file and returns
    {"lvl":lvl,"conf":conf,"term":term,"spin":spin,"l":l,"spd":spd,"j":j
    ,"mult":mult,"ecm":ecm,"eryd":eryd,"ecmth":ecmth,"erydth":erydth,"ref":ref,"pretty":pretty, 'ionS':ions}
    if a energy value for ecm or eryd is zero(=unknown), the theoretical values
    (ecmth and erydth) are inserted
    """
    #
    fstring='i3,i6,a15,i3,i3,a3,f4.1,i3,4f15.2'
    elvlcFormat=FortranFormat(fstring)
    #
    if type(filename) == NoneType:
        fname=ion2filename(ions)
        elvlname=fname+'.elvlc'
    else:
        elvlname = filename
        bname = os.path.basename(filename)
        ions = bname.split('.')[0]
    if not os.path.isfile(elvlname):
        print(' elvlc file does not exist:  ',elvlname)
        return {'status':0}
    status = 1
    input=open(elvlname,'r')
    s1=input.readlines()
    input.close()
    nlvls=0
    ndata=2
    while ndata > 1:
        s1a=s1[nlvls][:-1]
        s2=s1a.split()
        ndata=len(s2)
        nlvls=nlvls+1
    nlvls-=1
    if verbose:
        print(' nlvls = ', nlvls)
    lvl=[0]*nlvls
    conf=[0]*nlvls
    term=[0]*nlvls
    spin=[0]*nlvls
    l=[0]*nlvls
    spd=[0]*nlvls
    j=[0]*nlvls
    mult=[0]*nlvls
    ecm=[0]*nlvls
    eryd=[0]*nlvls
    ecmth=[0]*nlvls
    erydth=[0]*nlvls
    pretty=[0]*nlvls
    for i in range(0,nlvls):
        if verbose:
            print(s1[i][0:115])
        inpt=FortranLine(s1[i][0:115],elvlcFormat)
        lvl[i]=inpt[0]
        conf[i]=inpt[1]
        term[i]=inpt[2].strip()
        spin[i]=inpt[3]
        l[i]=inpt[4]
        spd[i]=inpt[5].strip()
        j[i]=inpt[6]
        mult[i]=inpt[7]
        ecm[i]=inpt[8]
        eryd[i]=inpt[9]
        ecmth[i]=inpt[10]
        erydth[i]=inpt[11]
        if ecm[i] == 0.:
            if useTh:
                ecm[i] = ecmth[i]
                eryd[i] = erydth[i]
        stuff = term[i].strip() + ' %1i%1s%3.1f'%( spin[i], spd[i], j[i])
        pretty[i] = stuff.strip()
    ref=[]
    for i in range(nlvls+1,len(s1)-1):
        s1a=s1[i][:-1]
        ref.append(s1a.strip())
#    self.const.Elvlc={"lvl":lvl,"conf":conf,"term":term,"spin":spin,"l":l,"spd":spd,"j":j
#            ,"mult":mult,"ecm":ecm,"eryd":eryd,"ecmth":ecmth,"erydth":erydth,"ref":ref}
    return {"lvl":lvl,"conf":conf,"term":term,"spin":spin,"l":l,"spd":spd,"j":j
            ,"mult":mult,"ecm":ecm,"eryd":eryd,"ecmth":ecmth,"erydth":erydth,"ref":ref,"pretty":pretty, 'ionS':ions, 'status':status}
    #
    # -------------------------------------------------------------------------------------
    #
def elvlcWrite(info):
    ''' creates a .elvlc in the current directory
    info is a dictionary that must contain the following keys
    ionS, the Chianti style name of the ion such as c_4
    conf, an integer denoting the configuration - not too essential
    term, a string showing the configuration
    spin, an integer of the spin of the state in LS coupling
    l, an integer of the angular momentum quantum number
    spd, an string for the alphabetic symbol of the angular momemtum, S, P, D, etc
    j, a floating point number, the total angular momentum
    ecm, the observed energy in inverse cm, if unknown, the value is 0.
    eryd, the observed energy in Rydbergs, if unknown, the value is 0.
    ecmth, the calculated energy from the scattering calculation, in inverse cm
    erydth, the calculated energy from the scattering calculation in Rydbergs
    ref, the references in the literature to the data in the input info'''
    gname = info['ionS']
    elvlcname = gname + '.elvlc'
    print(' elvlc file name = ', elvlcname)
    out = open(elvlcname, 'w')
    for i,  conf in enumerate(info['conf']):
        mult = int(2.*info['j'][i]+1.)
        pstring = '%3i%6s%15s%3i%3i%2s%5.1f%3i%15.3f%15.6f%15.3f%15.6f \n'%(i+1, conf, info['term'][i], info['spin'][i], info['l'][i], info['spd'][i], info['j'][i], mult, info['ecm'][i], info['eryd'][i], info['ecmth'][i], info['erydth'][i])
    #i3,a6,a15,2i3,a2,f5.1,i3,f15.3,f15.6,f15.3,f15.6
        out.write(pstring)
    out.write(' -1\n')
    out.write('%filename:  ' + elvlcname + '\n')
    info['ref'].append(' produced as a part of the George Mason University, University of Cambridge, University of Michigan \'CHIANTI\' atomic database for astrophysical spectroscopy consortium')
    today = date.today()
    info['ref'].append(' K. Dere (GMU) - ' + today.strftime('%Y %B %d'))
    for one in info['ref']:
        out.write(one+'\n')
    out.write(' -1\n')
    out.close()
    return
    #
    # -------------------------------------------------------------------------------------
    #
def wgfaRead(ions, filename=0, elvlcname=0, total=0):
    """
    reads chianti wgfa file and returns
    {"lvl1":lvl1,"lvl2":lvl2,"wvl":wvl,"gf":gf,"avalue":avalue,"ref":ref}
    if elvlcname is specified, the lsj term labels are returned as 'pretty1' and 'pretty2'
    """
    #
    if filename:
        wgfaname = filename
        if not filename:
            elvlcname = os.path.splitext(wgfaname)[0] + '.elvlc'
    else:
        fname=ion2filename(ions)
        wgfaname=fname+'.wgfa'
        elvlcname = fname + '.elvlc'
    #
    if os.path.isfile(elvlcname):
        elvlc = elvlcRead('', elvlcname)
    else:
        elvlc = 0
    input=open(wgfaname,'r')
    s1=input.readlines()
    dum=input.close()
    nwvl=0
    ndata=2
    while ndata > 1:
        s1a=s1[nwvl][:-1]
        s2=s1a.split()
        ndata=len(s2)
        nwvl=nwvl+1
    nwvl=nwvl-1
    lvl1=[0]*nwvl
    lvl2=[0]*nwvl
    wvl=[0.]*nwvl
    gf=[0.]*nwvl
    avalue=[0.]*nwvl
    if elvlcname:
        pretty1 = ['']*nwvl
        pretty2 = ['']*nwvl
    #
    wgfaFormat='(2i5,f15.3,2e15.3)'
    for i in range(nwvl):
        inpt=FortranLine(s1[i],wgfaFormat)
        lvl1[i]=inpt[0]
        lvl2[i]=inpt[1]
        wvl[i]=inpt[2]
        gf[i]=inpt[3]
        avalue[i]=inpt[4]
        if elvlcname:
            pretty1[i] = elvlc['pretty'][inpt[0] - 1]
            pretty2[i] = elvlc['pretty'][inpt[1] - 1]

    ref=[]
    for i in range(nwvl+1,len(s1)-1):
        s1a=s1[i][:-1]
        ref.append(s1a.strip())
    Wgfa={"lvl1":lvl1,"lvl2":lvl2,"wvl":wvl,"gf":gf,"avalue":avalue,"ref":ref, 'ionS':ions, 'filename':wgfaname}
    if total:
        avalueLvl = [0.]*max(lvl2)
        for iwvl in range(nwvl):
            avalueLvl[lvl2[iwvl] -1] += avalue[iwvl]
        Wgfa['avalueLvl'] = avalueLvl

    if elvlcname:
        Wgfa['pretty1'] = pretty1
        Wgfa['pretty2'] = pretty2
    return Wgfa
    #
    # --------------------------------------
    #
def wgfaWrite(info, outfile = 0, minBranch = 0.):
    '''
    to write a wgfa file
    info is a dictionary the contains the following elements
    ionS, the Chianti style name of the ion such as c_4 for C IV
    lvl1 - the lower level, the ground level is 1
    lvl2 - the upper level
    wvl - the wavelength in Angstroms
    gf - the weighted oscillator strength
    avalue - the A value
    pretty1 - descriptive text of the lower level (optional)
    pretty2 - descriptive text of the upper level (optiona)
    ref - reference text, a list of strings
    minBranch:  the transition must have a branching ratio greater than the specified to be written to the file
    '''
    #
#    gname = info['ionS']
    if outfile:
        wgfaname = outfile
    else:
        wgfaname = gname + '.wgfa'
    print((' wgfa file name = ', wgfaname))
    if minBranch > 0.:
        info['ref'].append(' minimum branching ratio = %10.2e'%(minBranch))
    out = open(wgfaname, 'w')
    ntrans = len(info['lvl1'])
    nlvl = max(info['lvl2'])
    totalAvalue = np.zeros(nlvl, 'float64')
    if 'pretty1' in info:
        pformat = '%5i%5i%15.4f%15.3e%15.3e%30s - %30s'
    else:
        pformat = '%5i%5i%15.4f%15.3e%15.3e'
    for itrans, avalue in enumerate(info['avalue']):
        # for autoionization transitions, lvl1 can be less than zero
        if abs(info['lvl1'][itrans]) > 0 and info['lvl2'][itrans] > 0:
            totalAvalue[info['lvl2'][itrans] -1] += avalue

    for itrans, avalue in enumerate(info['avalue']):
        if avalue > 0.:
            branch = avalue/totalAvalue[info['lvl2'][itrans] -1]
        else:
            branch = 0.
        if branch > minBranch and abs(info['lvl1'][itrans]) > 0 and info['lvl2'][itrans] > 0:
            if 'pretty1' in info:
                # generally only useful with NIST data
                if 'transType' in info:
                    if info['transType'][itrans] != '':
                        lbl2 = info['pretty2']+'  ' + info['transType'][itrans]
                else:
                    lbl2= info['pretty2'][itrans]
                pstring= pformat%(info['lvl1'][itrans], info['lvl2'][itrans], info['wvl'][itrans], info['gf'][itrans], avalue, info['pretty1'][itrans].rjust(30), lbl2.ljust(30))
                out.write(pstring+'\n')
            else:
                pstring= pformat%(info['lvl1'][itrans], info['lvl2'][itrans], info['wvl'][itrans], info['gf'][itrans], avalue)
                out.write(pstring+'\n')
    out.write(' -1\n')
    out.write('%filename:  ' + wgfaname + '\n')
    for one in info['ref']:
        out.write(one+'\n')
    out.write(' -1\n')
    out.close()
    #
    # -------------------------------------------------------------------------------------
    #
def easplomRead(ions, extension='.splom'):
    """read chianti splom files and returns
    {"lvl1":lvl1,"lvl2":lvl2,"deryd":de,"gf":gf,"eryd":eout,"omega":omout}
    currently only works for 5 point spline fit files"""
    #
    #
    fname=ion2filename(ions)
    omname=fname+extension
    input=open(omname,'r')
    lines=input.readlines()
    input.close()
    format=FortranFormat('5i3,8e10.3')
    data=5
    iline=0
    lvl1=[]
    lvl2=[]
    ttype=[]
    gf=[]
    de=[]
    om=[]
    z=1
    while z > 0:
        omdat1=FortranLine(lines[iline],format)
        z=omdat1[0]
        if z > 0:
            l1=omdat1[2]
            l2=omdat1[3]
            ttype1=omdat1[4]
            gf1=omdat1[5]
            de1=omdat1[6]
            btf1=omdat1[7]
            om1=omdat1[8:]
            #
            lvl1.append(l1)
            lvl2.append(l2)
            ttype.append(ttype1)
            gf.append(gf1)
            de.append(de1)
            om.append(om1)
        iline=iline+1
    omout=np.asarray(om,'Float64')
    ref=lines[iline:-1]
#        omout=np.transpose(omout)
    if extension == '.omdat':
        Splom={"lvl1":lvl1,"lvl2":lvl2,'ttype':ttype,"gf":gf, "deryd":de,"omega":omout, 'ref':ref}
        return Splom
    elif  extension == '.easplom':
        Easplom={"lvl1":lvl1,"lvl2":lvl2,'ttype':ttype,"gf":gf, "deryd":de,"omega":omout, 'ref':ref}
        return Easplom

    return
    #
    #-----------------------------------------------------------
    #
def splomDescale(splom, energy):
    """
    Calculates the collision strength
    for excitation-autoionization as a function of energy.
    energy in eV
    """
    #
    #
    nenergy=energy.size
    nsplom=len(splom['deryd'])
    # for these files, there are 5 spline points
    nspl = 5
    if nenergy > 1:
        omega = np.zeros((nsplom,nenergy),"float64")
    else:
        omega = np.zeros(nsplom,"float64")
    #
    dx = 1./(float(nspl)-1.)
    sxint = dx*np.arange(nspl)
    for isplom in range(0,nsplom):
        #
        sx1 = energy/(splom['deryd'][isplom]*const.ryd2Ev)
        good = sx1 >= 1.
        # make sure there are some valid energies above the threshold
        if good.sum():
            nbad = nenergy - good.sum()
            c_curr = splom['c'][isplom]
            #
            if splom['ttype'][isplom] == 1:
                sx = 1. - np.log(c_curr)/np.log(sx1[good] - 1. + c_curr)
                y2 = interpolate.splrep(sxint,splom['splom'][:, isplom],s=0)  #allow smoothing,s=0)
                som = interpolate.splev(sx,y2,der=0)
                omega[isplom, nbad:] = som*np.log(sx -1. + np.exp(1.))
            #
            elif splom['ttype'][isplom] == 2:
                sx =(sx1[good] - 1.)/(sx1[good] -1. + c_curr)
                y2 = interpolate.splrep(sxint,splom['splom'][:, isplom],s=0)  #allow smoothing,s=0)
                som=interpolate.splev(sx,y2,der=0)
                omega[isplom, nbad:] = som
            #
            elif splom['ttype'][isplom] == 3:
                sx = (sx1[good] - 1.)/(sx1[good] -1. + c_curr)
                y2 = interpolate.splrep(sxint,splom['splom'][:, isplom],s=0)  #allow smoothing,s=0)
                som = interpolate.splev(sx,y2,der=0)
                omega[isplom, nbad:] = som/sx1[good]**2
            #
            elif splom['ttype'][isplom] == 4:
                sx = 1. - np.log(c_curr)/np.log(sx1[good] -1. + c_curr)
                y2 = interpolate.splrep(sxint,splom['splom'][:, isplom],s=0)  #allow smoothing,s=0)
                som=interpolate.splev(sx,y2,der=0)
                omega[isplom, nbad:] = som*np.log(sx1[good] -1. + c_curr)
            #
            #
            #
            elif ttype > 4:
                print((' splom t_type ne 1,2,3,4 = %4i %4i %4i'%(ttype,l1,l2)))
        else:
            # there are no energies above the threshold
            pass
    #
    #
    omega=np.where(omega > 0.,omega,0.)
    #
    return omega
    #
    # --------------------------------------------------
    #
def splomRead(ions, ea=0, filename=None):
    """
    read chianti .splom files and return
    {"lvl1":lvl1,"lvl2":lvl2,"ttype":ttype,"gf":gf,"deryd":de,"c":c,"splom":splomout,"ref":hdr} not tested
    """
    #
    if type(filename) == NoneType:
        fname=ion2filename(ions)
        if ea:
            splomname=fname+'.easplom'
        else:
            splomname=fname+'.splom'
    else:
        splomname = filename
    input=open(splomname,'r')
    #  need to read first line and see how many elements
    line1=input.readline()
    indices=line1[0:15]
    remainder=line1[16:]
    nom=remainder.split(' ')
    format=FortranFormat('5i3,'+str(len(nom))+'E10.2')
    #  go back to the beginning
    input.seek(0)
    lines=input.readlines()
    data=5
    iline=0
    lvl1=[]
    lvl2=[]
    ttype=[]
    gf=[]
    de=[]
    f=[]
    splom=[]
    ntrans=0
    while data > 1:
        splomdat=FortranLine(lines[iline],format)
        l1=splomdat[2]
        l2=splomdat[3]
        tt1=splomdat[4]
        gf1=splomdat[5]
        de1=splomdat[6]
        f1=splomdat[7]
        splom1=splomdat[8:]
        lvl1.append(int(l1))
        lvl2.append(int(l2))
        ttype.append(int(tt1))
        gf.append(float(gf1))
        de.append(float(de1))
        f.append(float(f1))
        splom.append(splom1)
        iline=iline+1
        data=len(lines[iline].split(' ',2))
    hdr=lines[iline+1:-1]
    de=np.asarray(de,'Float64')
    splomout=np.asarray(splom,'Float64')
    splomout=np.transpose(splomout)
    input.close()
    # note:  de is in Rydbergs
    splom={"lvl1":lvl1,"lvl2":lvl2,"ttype":ttype,"gf":gf,"deryd":de,"c":f
        ,"splom":splomout,"ref":hdr}
    return  splom
    #
    # --------------------------------------------------
    #
def splupsRead(ions, filename=0, prot=0, ci=0,  diel=0):
    """
    read a chianti splups file and return
    {"lvl1":lvl1,"lvl2":lvl2,"ttype":ttype,"gf":gf,"de":de,"cups":cups,"bsplups":bsplups,"ref":ref}
    if prot >0, then reads the psplups file
    if ci > 0, then reads cisplups file
    if diel > 0, then reads dielsplups file
    """
    #
    if filename:
        splupsname = filename
    else:
        fname=ion2filename(ions)
        if prot:
            splupsname=fname+'.psplups'
        elif ci:
            splupsname=fname+'.cisplups'
        elif diel:
            splupsname=fname+'.dielsplups'
        else:
            splupsname=fname+'.splups'
    if not os.path.exists(splupsname):
        if prot:
            return None
        elif ci:
            return None
        else:
            return None
    # there is splups/psplups data
    else:
        input=open(splupsname,'r')
        s1=input.readlines()
        input.close()
        nsplups=0
        ndata=2
        while ndata > 1:
            s1a=s1[nsplups][:]
            s2=s1a.split()
            ndata=len(s2)
            nsplups=nsplups+1
        nsplups=nsplups-1
        lvl1=[0]*nsplups
        lvl2=[0]*nsplups
        ttype=[0]*nsplups
        gf = np.zeros(nsplups, 'float64')
        de= np.zeros(nsplups, 'float64')
        cups= np.zeros(nsplups, 'float64')
        nspl=[0]*nsplups
#        splups=np.zeros((nsplups,9),'Float64')
        splups = [0.]*nsplups
        if prot:
#            splupsFormat1 = FortranFormat('3i3,8e10.3')
            splupsFormat2 = FortranFormat('3i3,3e10.3')
        else:
#            splupsFormat1='(6x,3i3,8e10.3)'
            splupsFormat2 = FortranFormat('6x,3i3,3e10.3')
        #
        for i in range(0,nsplups):
            inpt=FortranLine(s1[i],splupsFormat2)
            lvl1[i]=inpt[0]
            lvl2[i]=inpt[1]
            ttype[i]=inpt[2]
            gf[i]=inpt[3]
            de[i]=inpt[4]
            cups[i]=inpt[5]
            if prot:
                as1 = s1[i][39:].rstrip()
            else:
                as1 = s1[i][45:].rstrip()
            nspl[i] = len(as1)/10
            splupsFormat3 = FortranFormat(str(nspl[i])+'E10.2')
#            splupsFormat3 = '(' + str(nspl[i]) + 'e10.3' + ')'
            inpt = FortranLine(as1, splupsFormat3)
            spl1 = np.asarray(inpt[:], 'float64')
            splups[i] = spl1
        #
        ref=[]
        for i in range(nsplups+1,len(s1)-1):
            s1a=s1[i][:-1]
            ref.append(s1a.strip())
        if prot:
#            self.Npsplups=nsplups
#            self.Psplups={"lvl1":lvl1,"lvl2":lvl2,"ttype":ttype,"gf":gf,"de":de,"cups":cups
#                ,"nspl":nspl,"splups":splups,"ref":ref}
            return {"lvl1":lvl1,"lvl2":lvl2,"ttype":ttype,"gf":gf,"de":de,"cups":cups
                ,"nspl":nspl,"splups":splups,"ref":ref, 'filename':splupsname}
        else:
#            self.Splups={"lvl1":lvl1,"lvl2":lvl2,"ttype":ttype,"gf":gf,"de":de,"cups":cups
#                ,"nspl":nspl,"splups":splups,"ref":ref}
            return {"lvl1":lvl1,"lvl2":lvl2,"ttype":ttype,"gf":gf,"de":de,"cups":cups
                ,"nspl":nspl,"splups":splups,"ref":ref, 'filename':splupsname}
    #
    # -------------------------------------------------------------------------------------
    #
def cireclvlRead(ions, filename=0, cilvl=0, reclvl=0, rrlvl=0):
    '''
    to read Chianti cilvl and reclvl files and return data
    must specify type as either cilvl, reclvl or rrlvl
    '''
    if filename:
        fname = filename
    else:
        fname = ion2filename(ions)
    if cilvl:
        paramname=fname+'.cilvl'
    elif reclvl:
        paramname = fname + '.reclvl'
    elif rrlvl:
        paramname = fname + '.rrlvl'
    else:
        print('either "cilvl", "reclvl" ir "rrlvl" must be specified')
        return {}
    if os.path.exists(paramname):
        input=open(paramname,'r')
        lines = input.readlines()
        input.close()
    else:
        print(('file does not exist:  ', paramname))
        return {'error':'file does not exist: ' + paramname}
    #
    iline = 0
    idx = -1
    while idx < 0:
        aline=lines[iline][0:5]
        idx=aline.find('-1')
        iline += 1
    ndata = iline - 1
    ntrans = ndata/2
    #
    nref = 0
    idx = -1
    while idx < 0:
        aline=lines[iline][0:5]
        idx=aline.find('-1')
        iline += 1
        nref += 1
    nref -= 1
    #
    # need to find the maximum number of temperatures, not all lines are the same
    #
    ntemp = np.zeros(ntrans, 'int32')
    iline = 0
    for jline in range(0, ndata, 2):
        dummy = lines[jline].replace(os.linesep, '').split()
        ntemp[iline] = len(dummy[4:])
        iline += 1
    maxNtemp = ntemp.max()
#   print ' maxNtemp = ', maxNtemp
    temp = np.zeros((ntrans,maxNtemp), 'float64')
    iline = 0
    for jline in range(0, ndata, 2):
        recdat = lines[jline].replace(os.linesep, '').split()
        shortT = np.asarray(recdat[4:], 'float64')
        # the result of the next statement is to continue to replicate t
        t = np.resize(shortT, maxNtemp)
        if rrlvl:
            temp[iline] = t
        else:
            temp[iline] = 10.**t
        iline += 1
    #
    lvl1 = np.zeros(ntrans, 'int64')
    lvl2 = np.zeros(ntrans, 'int64')
    ci = np.zeros((ntrans, maxNtemp), 'float64')
    #
    idat = 0
    for jline in range(1, ndata, 2):
        cidat = lines[jline].replace(os.linesep, '').split()
        shortCi = np.asarray(cidat[4:], 'float64')
        lvl1[idat] = int(cidat[2])
        lvl2[idat] = int(cidat[3])
        ci[idat] = np.resize(shortCi, maxNtemp)
        idat += 1
    return {'temperature':temp, 'ntemp':ntemp,'lvl1':lvl1, 'lvl2':lvl2, 'rate':ci,'ref':lines[ndata+1:-1], 'ionS':ions}
    #
    # ----------------------------------------------------------
    #
def dilute(radius):
    '''
    to calculate the dilution factor as a function distance from
    the center of a star in units of the stellar radius
    a radius of less than 1.0 (incorrect) results in a dilution factor of 0.
    '''
    if radius >= 1.:
        d = 0.5*(1. - np.sqrt(1. - 1./radius**2))
    else:
        d = 0.
    return d
    #
    #
    # ------------------------------------------------------------------------------
    #
    #
    #-----------------------------------------------------------
    #
def diRead(ions, filename=0):
    """
    read chianti direct ionization .params files and return
        {"info":info,"btf":btf,"ev1":ev1,"xsplom":xsplom,"ysplom":ysplom,"ref":hdr}
        info={"iz":iz,"ion":ion,"nspl":nspl,"neaev":neaev}
    """
    #
    if filename:
        paramname = filename
    else:
        zion=convertName(ions)
        if zion['Z'] < zion['Ion']:
            print(' this is a bare nucleus that has no ionization rate')
            return
        #
        fname=ion2filename(ions)
        paramname=fname+'.diparams'
    #
    input=open(paramname,'r')
    #  need to read first line and see how many elements
    line1=input.readline()
    indices=line1.split()
    iz=int(indices[0])
    ion=int(indices[1])
    nspl=indices[2]
    nfac=int(indices[3])
    neaev=int(indices[4])
    nspl=int(nspl)
    format=FortranFormat(str(nspl+1)+'E10.2')
    #
    ev1=np.zeros(nfac,'Float64')
    btf=np.zeros(nfac,'Float64')
    xsplom=np.zeros([nfac, nspl],'Float64')
    ysplom=np.zeros([nfac, nspl],'Float64')
    #
    for ifac in range(nfac):
        line=input.readline()
        paramdat=FortranLine(line,format)
        btf[ifac]=paramdat[0]
        xsplom[ifac]=paramdat[1:]
        line=input.readline()
        paramdat=FortranLine(line,format)
        ev1[ifac]=paramdat[0]
        ysplom[ifac]=paramdat[1:]
    if neaev:
        line=input.readline()
        eacoef=line.split()
#            print ' eaev = ', type(eacoef), eacoef
        eaev=[float(avalue) for avalue in eacoef]
#            print ' eaev = ', type(eaev), eaev
#            print ' eaev = ', type(eaev), eaev
#            if len(eaev) == 1:
#                eaev=float(eaev[0])
#                eaev=np.asarray(eaev, 'float32')
#            else:
#                eaev=np.asarray(eaev, 'float32')
    else:
        eaev=0.
    hdr=input.readlines()
    input.close()
    info={"iz":iz,"ion":ion,"nspl":nspl,"neaev":neaev, 'nfac':nfac}
    if neaev:
        info['eaev'] = eaev
    DiParams={"info":info,"btf":btf,"ev1":ev1,"xsplom":xsplom,"ysplom":ysplom, 'eaev':eaev,"ref":hdr}
    return DiParams
    #
    # -------------------------------------------------------------------------------------
    #
def eaCross(diparams, easplom, elvlc, energy=None, verbose=False):
    '''
    Provide the excitation-autoionization cross section.

    Energy is given in eV.
    '''
    energy = np.asarray(energy, 'float64')
    if not energy.any():
        btenergy=0.1*np.arange(10)
        btenergy[0]=0.01
        dum=np.ones(len(btenergy))
        [energy, dum] = descale_bti(btenergy, dum, 2., min(easplom['deryd']))
        energy=np.asarray(energy, 'float64')
    #
    omega = splomDescale(easplom, energy)
    #
    #  need to replicate neaev

    if diparams['info']['neaev'] > 0:
        f1 = np.ones(omega.shape[0])
    else:
        f1 = diparams['info']['eaev']

    totalCross = np.zeros_like(energy)
    ntrans = omega.shape[0]
    for itrans in range(ntrans):
        lvl1 = easplom['lvl1'][itrans]
        mult = 2.*elvlc['j'][lvl1 - 1] + 1.
        cross = f1[itrans]*const.bohrCross*omega[itrans]/(mult.energy/const.ryd2Ev)
        totalCross += cross
    return {'energy':energy, 'cross':totalCross}
    #
    # -------------------------------------------------------------------------------------
    #
def eaRead(ions, filename=0):
    '''
    read a chianti excitation-autoionization file and return the EA ionization rate data
    derived from splupsRead
    {"lvl1":lvl1,"lvl2":lvl2,"ttype":ttype,"gf":gf,"de":de,"cups":cups,"bsplups":bsplups,"ref":ref}
    '''
    if filename:
        splupsname = filename
    else:
        zion=convertName(ions)
        if zion['Z'] < zion['Ion']:
            print(' this is a bare nucleus that has no ionization rate')
            return
        #
        fname=ion2filename(ions)
        splupsname=fname+'.easplups'
    if not os.path.exists(splupsname):
        print((' could not find file:  ', splupsname))
        Splups={"lvl1":-1}
        return {"lvl1":-1}
    # there is splups/psplups data
    else:
        input=open(splupsname,'r')
        s1=input.readlines()
        dum=input.close()
        nsplups=0
        ndata=2
        while ndata > 1:
            s1a=s1[nsplups][:]
            s2=s1a.split()
            ndata=len(s2)
            nsplups=nsplups+1
        nsplups=nsplups-1
        lvl1=[0]*nsplups
        lvl2=[0]*nsplups
        ttype=[0]*nsplups
        gf=[0.]*nsplups
        de=[0.]*nsplups
        cups=[0.]*nsplups
        nspl=[0]*nsplups
        splups=np.zeros((nsplups,9),'Float64')
        splupsFormat1='(6x,3i3,8e10.0)'
        splupsFormat2='(6x,3i3,12e10.0)'
        #
        for i in range(0,nsplups):
            try:
                inpt=FortranLine(s1[i],splupsFormat1)
            except:
                inpt=FortranLine(s1[i],splupsFormat2)
            lvl1[i]=inpt[0]
            lvl2[i]=inpt[1]
            ttype[i]=inpt[2]
            gf[i]=inpt[3]
            de[i]=inpt[4]
            cups[i]=inpt[5]
            if len(inpt)  > 13:
                nspl[i]=9
                splups[i].put(range(9),inpt[6:])
            else:
                nspl[i]=5
                splups[i].put(range(5),inpt[6:])
        #
        ref=[]
        for i in range(nsplups+1,len(s1)-1):
            s1a=s1[i][:-1]
            ref.append(s1a.strip())
#        self.EaParams={"lvl1":lvl1,"lvl2":lvl2,"ttype":ttype,"gf":gf,"de":de,"cups":cups
#                ,"nspl":nspl,"splups":splups,"ref":ref}
        return {"lvl1":lvl1,"lvl2":lvl2,"ttype":ttype,"gf":gf,"de":de,"cups":cups
                ,"nspl":nspl,"splups":splups,"ref":ref}
    #
    # -------------------------------------------------------------------------------------
    #
def rrRead(ions):
    """read chianti radiative recombination .rrparams files and return
        {'rrtype','params','ref'}"""
    #
    #
    fname=ion2filename(ions)
    paramname=fname+'.rrparams'
    if os.path.isfile(paramname):
        input=open(paramname,'r')
        #  need to read first line and see how many elements
        lines=input.readlines()
        input.close()
        rrtype=int(lines[0])
        ref=lines[3:-2]
        #
        if rrtype == 1:
            # a Badnell type
            fmt=FortranFormat('3i5,e12.4,f10.5,2e12.4')
            params=FortranLine(lines[1],fmt)
            RrParams={'rrtype':rrtype, 'params':params, 'ref':ref}
        elif rrtype == 2:
            # a Badnell type
            fmt=FortranFormat('3i5,e12.4,f10.5,2e11.4,f10.5,e12.4')
            params=FortranLine(lines[1],fmt)
            RrParams={'rrtype':rrtype, 'params':params, 'ref':ref}
        elif rrtype == 3:
            # a Shull type
            fmt=FortranFormat('2i5,2e12.4')
            params=FortranLine(lines[1],fmt)
            RrParams={'rrtype':rrtype, 'params':params, 'ref':ref}
        else:
            RrParams=None
            print((' for ion %5s unknown RR type = %5i' %(ions, rrtype)))
        return RrParams
    else:
        return {'rrtype':-1}

    #
    # -------------------------------------------------------------------------------------
    #
def drRead(ions):
    """read chianti dielectronic recombination .drparams files and return
        {'rrtype','params','ref'}"""
    #
    #
    fname=ion2filename(ions)
    paramname=fname+'.drparams'
    if os.path.isfile(paramname):
        input=open(paramname,'r')
        #  need to read first line and see how many elements
        lines=input.readlines()
        input.close()
        drtype=int(lines[0])
        ref=lines[4:-1]
        #
        if drtype == 1:
            # a Badnell type
            fmt=FortranFormat('2i5,8e12.4')
            eparams=np.asarray(FortranLine(lines[1],fmt)[2:], 'float64')
            cparams=np.asarray(FortranLine(lines[2],fmt)[2:], 'float64')
            DrParams={'drtype':drtype, 'eparams':eparams,'cparams':cparams,  'ref':ref}
        elif drtype == 2:
            # shull type
            fmt=FortranFormat('2i5,4e12.4')
            params=np.asarray(FortranLine(lines[1],fmt)[2:], 'float64')
            DrParams={'drtype':drtype, 'params':params, 'ref':ref}
        else:
            DrParams = None
            print((' for ion %5s unknown DR type = %5i' %(ions, drtype)))
    else:
        DrParams=None
    return DrParams
    #
    # -------------------------------------------------------------------------------------
    #
def ioneqRead(ioneqname='', verbose=0):
    """
    reads an ioneq file and stores temperatures and ionization
    equilibrium values in self.IoneqTemperature and self.Ioneq and returns
    a dictionary containing these value and the reference to the literature
    """
    pass
    #
    # -------------------------------------------------------------------------------------
    #
def gffRead():
    '''to read the free-free gaunt factors of Sutherland, 1998, MNRAS, 300, 321.
    this function reads the file and reverses the values of g2 and u'''
    xuvtop = os.environ['XUVTOP']
    fileName = os.path.join(xuvtop, 'continuum','gffgu.dat' )
    input = open(fileName)
    lines = input.readlines()
    input.close()
    #
    #  the 1d stuff below is to make it easy to use interp2d
    ngamma=41
    nu=81
    nvalues = ngamma*nu
    g2 = np.zeros(ngamma, 'float64')
    g21d = np.zeros(nvalues, 'float64')
    u = np.zeros(nu, 'float64')
    u1d = np.zeros(nvalues, 'float64')
    gff = np.zeros((ngamma, nu), 'float64')
    gff1d = np.zeros(nvalues, 'float64')
    #
    iline = 5
    ivalue = 0
    for ig2 in range(ngamma):
        for iu in range(nu):
            values = lines[iline].split()
            u[iu] = float(values[1])
            u1d[ivalue] = float(values[1])
            g2[ig2] = float(values[0])
            g21d[ivalue] = float(values[0])
            gff[ig2, iu] = float(values[2])
            gff1d[ivalue] = float(values[2])
            iline+=1
            ivalue += 1
    #
    return {'g2':g2, 'g21d':g21d,  'u':u, 'u1d':u1d,  'gff':gff,  'gff1d':gff1d}
    #
    # ----------------------------------------------------------------------------------------
    #
def gffintRead():
    '''to read the integrated free-free gaunt factors of Sutherland, 1998, MNRAS, 300, 321.'''
    xuvtop = os.environ['XUVTOP']
    fileName = os.path.join(xuvtop, 'continuum','gffint.dat' )
    input = open(fileName)
    lines = input.readlines()
    input.close()
    #
    ngamma=41
    g2 = np.zeros(ngamma, 'float64')
    gffint = np.zeros(ngamma, 'float64')
    s1 = np.zeros(ngamma, 'float64')
    s2 = np.zeros(ngamma, 'float64')
    s3 = np.zeros(ngamma, 'float64')
    #
    ivalue = 0
    start = 4
    for iline in range(start,start+ngamma):
        values = lines[iline].split()
        g2[ivalue] = float(values[0])
        gffint[ivalue] = float(values[1])
        s1[ivalue] = float(values[2])
        s2[ivalue] = float(values[3])
        s3[ivalue] = float(values[4])
        ivalue += 1
    #
    return {'g2':g2, 'gffint':gffint, 's1':s1, 's2':s2, 's3':s3}
    #
    # ----------------------------------------------------------------------------------------
    #
def itohRead():
    '''to read in the free-free gaunt factors of Itoh et al. (ApJS 128, 125, 2000)'''
    xuvtop = os.environ['XUVTOP']
    itohName = os.path.join(xuvtop, 'continuum', 'itoh.dat')
    input = open(itohName)
    lines = input.readlines()
    input.close()
    gff = np.zeros((30, 121), 'float64')
    for iline in range(30):
        gff[iline]= np.asarray(lines[iline].split(), 'float64')
    return {'itohCoef':gff}
    #
    #
    # ----------------------------------------------------------------------------------------
    #
def klgfbRead():
    '''to read CHIANTI files file containing the free-bound gaunt factors for n=1-6 from Karzas and Latter, 1961, ApJSS, 6, 167
    returns {pe, klgfb}, the photon energy and the free-bound gaunt factors'''
    xuvtop = os.environ['XUVTOP']
    fname = os.path.join(xuvtop, 'continuum', 'klgfb.dat')
    input = open(fname)
    lines = input.readlines()
    input.close()
    #
    ngfb = int(lines[0].split()[0])
    nume = int(lines[0].split()[1])

    gfb = np.zeros((ngfb, ngfb, nume), 'float64')
    nlines = len(lines)
#        print 'nlines, nume, ngfb = ', nlines,  nume, ngfb
    pe = np.asarray(lines[1].split(), 'float64')
    for iline in range(2, nlines):
        data = lines[iline].split()
        n = int(data[0])
        l = int(data[1])
        gfb[n-1, l] = np.array(data[2:], 'float64')
    return {'pe':pe, 'klgfb':gfb}
#
#  ---------------------------------------------------------
#
def listFiles(path):
    '''
    walks the path and subdirectories to return a list of files
    '''
    alist=os.walk(path)
    print(' getting file list')
    listname=[]
    for (dirpath,dirnames,filenames) in alist:
        if len(dirnames) == 0:
            for f in filenames:
                file=os.path.join(dirpath,f)
                if os.path.isfile(file):
                    listname.append(file)
        else:
            for f in filenames:
                file=os.path.join(dirpath,f)
                if os.path.isfile(file):
                    listname.append(file)
    return listname
    #
    # ----------------------------------------------------------------------------------------
    #
def fblvlRead(filename, verbose=False):
    """ read a chianti energy level file and returns
    {"lvl":lvl,"conf":conf,"term":term,"spin":spin,"l":l,"spd":spd,"j":j
    ,"mult":mult,"ecm":ecm,"eryd":eryd,"ref":ref}"""
#        #  ,format='(i5,a20,2i5,a3,i5,2f20.3)'
    fstring='i5,a20,2i5,a3,i5,2f20.3'
    elvlcFormat=FortranFormat(fstring)
    #
    if os.path.exists(filename):
        input=open(filename,'r')
        s1=input.readlines()
        input.close()
        nlvls=0
        ndata=2
        while ndata > 1:
            s1a=s1[nlvls][:-1]
            s2=s1a.split()
            ndata=len(s2)
            nlvls=nlvls+1
        nlvls-=1
        if verbose:
            print((' nlvls = %5i'%(nlvls)))
        lvl=[0]*nlvls
        conf=[0]*nlvls
        pqn=[0]*nlvls
        l=[0]*nlvls
        spd=[0]*nlvls
        mult=[0]*nlvls
        ecm=[0]*nlvls
        ecmth=[0]*nlvls
        for i in range(0,nlvls):
            if verbose:
                print((s1[i]))
            inpt=FortranLine(s1[i],elvlcFormat)
            lvl[i]=inpt[0]
            conf[i]=inpt[1].strip()
            pqn[i]=inpt[2]
            l[i]=inpt[3]
            spd[i]=inpt[4].strip()
            mult[i]=inpt[5]
            if inpt[6] == 0.:
                ecm[i]=inpt[7]
            else:
                ecm[i]=inpt[6]
                ecmth[i]=inpt[7]
        ref=[]
        for i in range(nlvls+1,len(s1)-1):
            s1a=s1[i][:-1]
            ref.append(s1a.strip())
        return {"lvl":lvl,"conf":conf,'pqn':pqn,"l":l,"spd":spd,"mult":mult,
            "ecm":ecm,'ecmth':ecmth, 'ref':ref}
    else:
        return {'errorMessage':' fblvl file does not exist'}
    #
    # -----------------------------------------------------------------
    #
def vernerRead():
    '''Reads the Verner & Yakovlev (A&AS 109, 125, 1995) photoionization cross-section data'''
    xuvtop = os.environ['XUVTOP']
    fname = os.path.join(xuvtop, 'continuum', 'verner_short.txt')
    input = open(fname)
    lines = input.readlines()
    input.close()
    #
    nlines=465
    maxZ = 30+1
    maxNel = 30 +1# also equal max(stage)
    #
    #z = np.array(nlines,'int32')
    #nel = np.array(nlines,'int32')
    pqn = np.zeros((maxZ,maxNel),'int32')
    l = np.zeros((maxZ,maxNel),'int32')
    eth = np.zeros((maxZ,maxNel),'float64')
    e0 = np.zeros((maxZ,maxNel),'float64')
    sig0 = np.zeros((maxZ,maxNel),'float64')
    ya = np.zeros((maxZ,maxNel),'float64')
    p = np.zeros((maxZ,maxNel),'float64')
    yw = np.zeros((maxZ,maxNel),'float64')
    #
    fstring='i2,i3,i2,i2,6f11.3'
    vernerFormat=FortranFormat(fstring)
    #
    for iline in range(nlines):
        out=FortranLine(lines[iline],vernerFormat)
        z = out[0]
        nel = out[1]
        stage = z - nel + 1
        pqn[z,stage] = out[2]
        l[z,stage] = out[3]
        eth[z,stage] = out[4]
        e0[z,stage] = out[5]
        sig0[z,stage] = out[6]
        ya[z,stage] = out[7]
        p[z,stage] = out[8]
        yw[z,stage] = out[9]
    #
    return {'pqn':pqn, 'l':l, 'eth':eth, 'e0':e0, 'sig0':sig0, 'ya':ya, 'p':p, 'yw':yw}
    #
    #-----------------------------------------------------------
    #
def twophotonHRead():
    ''' to read the two-photon A values and distribution function for the H seq'''
    xuvtop = os.environ['XUVTOP']
    fName = os.path.join(xuvtop, 'continuum', 'hseq_2photon.dat')
    dFile = open(fName, 'r')
    a = dFile.readline()
    y0 = np.asarray(a.split())
    a = dFile.readline()
    z0 = np.asarray(a.split())
    nz = 30
    avalue = np.zeros(nz, 'float64')
    asum = np.zeros(nz, 'float64')
    psi0 = np.zeros((nz, 17), 'float64')
    for iz in range(nz):
        a=dFile.readline().split()
        avalue[iz] = float(a[1])
        asum[iz] = float(a[2])
        psi = np.asarray(a[3:])
        psi0[iz] = psi
    dFile.close()
    return {'y0':y0, 'z0':z0, 'avalue':avalue, 'asum':asum, 'psi0':psi0.reshape(30, 17)}
    #
    #-----------------------------------------------------------
    #
def twophotonHeRead():
    ''' to read the two-photon A values and distribution function for the He seq'''
    xuvtop = os.environ['XUVTOP']
    fName = os.path.join(xuvtop, 'continuum', 'heseq_2photon.dat')
    dFile = open(fName, 'r')
    a = dFile.readline()
    y0 = np.asarray(a.split())
    nz = 30
    avalue = np.zeros(nz, 'float64')
    psi0 = np.zeros((nz, 41), 'float64')
    for iz in range(1, nz):
        a=dFile.readline().split()
        avalue[iz] = float(a[1])
        psi = np.asarray(a[2:])
        psi0[iz] = psi
    dFile.close()
    return {'y0':y0, 'avalue':avalue, 'psi0':psi0.reshape(30, 41)}
    #
    #-----------------------------------------------------------
    #
def upsdatRead(upsdatFileName):
    '''
    to read the standard Chianti upsdat file
    '''
    inpt = open(upsdatFileName)
    lines = inpt.readlines()
    inpt.close()
    nTemp = int(lines[0].strip('\n'))
    minusOne = 0
    counter = 1
    ll = lines[1].split()
    temp = np.asarray(ll[3:], 'float64')
    while not minusOne:
        if '-1' in lines[counter][:4]:
            minusOne = 1
        else:
            counter += 1
    ntrans = (counter)/2
    lvl1 = []
    lvl2 = []
    de = []
    gf = []
    upsilon = []
    counter = 1
    for itrans in range(ntrans):
        ll1 = lines[counter].split()
        lvl1.append(int(ll1[0]))
        lvl2.append(int(ll1[1]))
        de.append(float(ll1[2]))
        ll2 = lines[counter+1].split()
#        print ' ll2 = ', ll2
#        print ' gf = ', ll2[2]
        gf.append(float(ll2[2]))
        upsilon.append(np.asarray(ll2[3:], 'float64'))
        counter += 2
    counter += 1
    ref = []
    for aline in lines[counter:-1]:
        ref.append(aline.strip('\n'))
    return {'lvl1':lvl1, 'lvl2':lvl2, 'de':de, 'gf':gf, 'upsilon':upsilon, 'temperature':temp, 'ref':ref}
    #
    # -----------------------------------------------------
    #
def versionRead():
    """ read the version number of the CHIANTI database"""
    xuvtop = os.environ['XUVTOP']
    vFileName = os.path.join(xuvtop, 'VERSION')
    vFile = open(vFileName)
    versionStr = vFile.readline()
    vFile.close()
    return versionStr.strip()
    #
    #-----------------------------------------------------------
    #
def scale_bti(evin,crossin,f,ev1):
    """
    apply BT ionization scaling to (energy, cross-section)
    returns [bte,btx]
    """
    u=evin/ev1
    bte=1.-np.log(f)/np.log(u-1.+f)
    btx=u*crossin*(ev1**2)/(np.log(u)+1.)
    return [bte,btx]
    #
    #-----------------------------------------------------------
    #
def descale_bti(bte,btx,f,ev1):
    """
    descale BT ionization scaling
    returns [energy,cross-section]
    """
    u=1.-f+np.exp(np.log(f)/(1.-bte))
    energy=u*ev1
    cross=(np.log(u)+1.)*btx/(u*ev1**2)
    return [energy,cross]
    #
    #-----------------------------------------------------------
    #
def descale_bt(bte,btomega,f,ev1):
    """
    descale BT excitation scaling
    returns [energy,collision strength]
    """
    u=1.-f+np.exp(np.log(f)/(1.-bte))
    energy=u*ev1
    omega=(np.log(u)-1.+np.exp(1.))*btomega
    return [energy,omega]
    #
    #-----------------------------------------------------------
    #
def scale_bt(evin,omega,f,ev1):
    """
    apply BT excitation scaling to (energy, collision strength)
    returns [bte,btomega]
    """
    u=evin/ev1
    bte=1.-np.log(f)/np.log(u-1.+f)
    btomega=omega/(np.log(u)-1.+np.exp(1.))
    return [bte,btomega]



"""
PyNeb - python package for the analysis of emission lines
"""

__all__ = []
from .version import __version__

import sys
__pyversion__ = sys.version_info[0]

from .utils.Config import _Config
config = _Config()
log_ = config.log_
log_.message('Starting PyNeb version %s' % __version__, calling='PyNeb')


if sys.version_info[0:2] < (2, 6):
    log_.warn('Python version >= 2.6 needed, seems you have {0}'.format(sys.version_info), calling='PyNeb')
try:
    from numpy.version import version as numpy_version    
    if [int(n) for n in (numpy_version.split('.')[:3])] < [1, 5, 1] :
        log_.warn('Numpy version >= 1.5.1 needed, seems you have {0}'.format(numpy_version), calling='PyNeb')
except:
    log_.warn('I do not understand what is your version of numpy: {0}, report this to PyNeb group'.format(numpy_version), calling='PyNeb')

from .utils.manage_atomic_data import _ManageAtomicData
atomicData = _ManageAtomicData()

atomicData.defaultDict = 'PYNEB_18_01'
atomicData.resetDataFileDict()

from .core.pynebcore import Atom, RecAtom, getAtomDict, getRecEmissivity, EmissionLine, Observation, \
    parseLineLabel, isValid
from .core.icf import ICF 
from .core.diags import Diagnostics, diags_dict
from .core.emisGrid import EmisGrid, getEmisGridDict
from .core.continuum import Continuum
from .plot.plotAtomicData import DataPlot
from .utils.physics import CST, print_IPs
from .utils.saverestore import save, restore
from .utils.init import LINE_LABEL_LIST, BLEND_LIST
from .utils.manage_atomic_data import getLevelsNIST
from .utils.misc import ROOT_DIR
from .extinction.red_corr import RedCorr

log_.message('PyNeb ready.', calling='PyNeb')
log_.timer('Starting PyNeb version %s' % __version__, quiet=True, calling='PyNeb')

__info__ = """PyNeb version: {}.
PyNeb is cited using: Luridiana, V.; Morisset, C.; Shaw, R. A. 2015, A&A, 573, 42
bibcode: 2015A&A...573A..42L
Default data dictionary: {}. This set of data can be changed by the user. 
Do not forget to cite the papers corresponding to the atomic data you are using.
PyNeb website: https://pypi.python.org/pypi/PyNeb
PyNeb discussion and hotline group email: pyneb@googlegroups.com
""".format(__version__, atomicData.defaultDict)



from datetime import datetime
from time import time

class PyNebError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)    
    
class my_logging(object):
    """
    Logging object that can log events with 3 different types (messages, warnings and error).
    Date and time are save as well as the calling procedure.
    events are printed in the standard output depending on the verbosity level.
    level: -1: Quiet and no log into log_ variable
             0: Quiet
             1: Only Errors
             2: Errors and Warnings
             3: very verbose, print Errors, Warnings and Messages 
             4: debug
    
    """
    
    def __init__(self, level=None, calling=None, file_=None, print_time=False, no_exit=True):
        """
        Parameters:
            - level [int] 0, 1, 2, 3, or 4
            - calling [str] description of the calling module
            - file_ [str] file to store the events (default is None).
            - print_time [boolean] if True, the event is printed with date-time (Default is False).

        """
        self.log_ = {}
        self.log_['messages'] = []
        self.log_['warnings'] = []
        self.log_['errors'] = []
        self.log_['timer'] = []
        self.log_['debug'] = []
        
        self.caller_max_size = 0
        self.level = level if level is not None else 2
        self.__start = time()
        if calling is not None:
            self.calling = calling
        else:
            self.calling = 'None'
        self.print_time = print_time
        self.open_file(file_)
        self.no_exit = no_exit
    
    def _pprint(self, mess_type, message, calling=None):
        if calling is None:
            calling = self.calling
        if self.print_time:
            print('{0} {1} @ {2}: {3}'.format(mess_type, calling, str(datetime.now()), message))
        else:
            print('{0} {1}: {2}'.format(mess_type, calling, message))
        
    def _add2log(self, mess_type, message, calling=None):
        if calling is None:
            calling = self.calling
        if self.level > -1:
            self.log_[mess_type].append((calling, str(datetime.now()), message))
            if len(calling) > self.caller_max_size:
                self.caller_max_size = len(calling)
            if self._to_file:
                self._file_out.write("{0} - {1} @ {2} : {3} \n".format(mess_type, calling, str(datetime.now()), message))

    def debug(self, message, calling=None):
        """
        method to add a debug message to the log                
        param:
            message [str] string to log
        """
        if self.level > 3:
            self._pprint('debug', message, calling=calling)
        self._add2log('debug', message, calling=calling)

    def message(self, message, calling=None):
        """
        method to add a message to the log
        param:
            message [str] string to log
        """
        if self.level > 2:
            self._pprint('    ', message, calling=calling)
        self._add2log('messages', message, calling=calling)

    def warn(self, message, calling=None):
        """
        method to add a warning to the log                
        param:
            message [str] string to log
        """
        if self.level > 1:
            self._pprint('warng', message, calling=calling)
        self._add2log('warnings', message, calling=calling)
        
    def error(self, message, calling=None, exception=None):
        """
        method to add an error to the log
        param:
            message [str] string to log
        """
        if self.level > 0:
            self._pprint('ERROR', message, calling=calling)
        self._add2log('errors', message, calling=calling)
        if exception is None:
            if self.no_exit:
                exception = PyNebError
            else:
                exception = SystemExit
        raise exception(message)
              
    def timer(self, message, quiet=False, calling=None):
        """
        method to add a timer to the log. Print the time elapsed since last call of timer (or instanciation
        of the object if first call).
        param:
            message [str] string to log
            quiet [boolean] if True, no message printed (but timer reset)
        """
        if calling is None:
            calling = self.calling
        delta_t = str((time() - self.__start))
        if not quiet:
            print('   {0}: {1} in {2}'.format(calling, message, delta_t))
        self.log_['timer'].append((calling, delta_t, message))
        if self._to_file:
            self._file_out.write("timer - {0} @ {1}: {2} \n".format(calling, delta_t, message))
        self.__start = time()

    def print_messages(self):
        """
        print all the messages.
        """
        to_print = '{0[0]:' + str(self.caller_max_size) + 's} at {0[1]}: {0[2]}' 
        for message in self.log_['messages']:
            print(to_print.format(message))
    
    def print_warnings(self):
        """
        print all the warnings.
        """
        to_print = '{0[0]:' + str(self.caller_max_size) + 's} at {0[1]}: {0[2]}' 
        for message in self.log_['warnings']:
            print(to_print.format(message))
    
    def print_errors(self):
        """
        print all the errors.
        """
        to_print = '{0[0]:' + str(self.caller_max_size) + 's} at {0[1]}: {0[2]}' 
        for message in self.log_['errors']:
            print(to_print.format(message))
    
    def print_debug(self):
        """
        print all the debug information.
        """
        to_print = '{0[0]:' + str(self.caller_max_size) + 's} at {0[1]}: {0[2]}' 
        for message in self.log_['debug']:
            print(to_print.format(message))
    
    def print_timer(self):
        """
        print the timer messages.
        """
        print(self.log_['timer'])
        
    def open_file(self, file_):
        """
        Open the file for the output and allow printing to it.
        param:
            file_ [str] file to save the log events.
        """
        if file_ is not None:
            self._file_out = open(file_, 'w')
            self._to_file = True
        else:
            self._to_file = False

    def close_file(self):
        """
        Close the output file and avoid printing to it.
        """
        if self._to_file:
            self._file_out.close()
            self._to_file = False




#!/usr/bin/env python

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

from pyneb.version import __version__

setup(name='PyNeb', 
      version=__version__,
      description='Nebular tools',
      author='Christophe Morisset, Valentina Luridiana',
      long_description=long_description,
      author_email='chris.morisset@gmail.com',
      url='http://www.iac.es/proyecto/PyNeb/',
      py_modules=['pyneb.test.test_pyneb'],
      packages=['pyneb',
		'pyneb.core',
		'pyneb.plot',
		'pyneb.utils',
		'pyneb.extinction'],
      package_data={'pyneb':['atomic_data_fits/*.fits',
			     'atomic_data_fits/*.hdf5',
			     'atomic_data_fits/old_fits/*.fits',
			     'atomic_data/*.dat',
			     'atomic_data/*.func',
			     'atomic_data/*.txt',
			     'atomic_data/levels/*.dat'],
                    'pyneb.extinction':['*.txt'],
                    'pyneb.plot':['level_diagrams/*.png']
                    }
     )




import pickle

def save(file_, *args, **kwargs):
    """
    Save the value of some data in a file.
    Usage: save('misdatos.pypic','a',b=b)

    """
    f=open(file_, "wb")
    dict = kwargs
    for name in args:
        dict[name] = eval(name)
    pickle.dump(dict, f, protocol=2)
    f.close

def restore(file_):
    """
    Read data saved with save function.
    Usage: datos = restore('misdatos.pypic')

    """
    f=open(file_, "rb")
    result = pickle.load(f)
    f.close
    return result



import numpy as np
import pandas as pd
import requests

def download_file(ion):
    """
    Download the file corresponding to the Mao, Badnell & Del Zanna 2020 paper on
    R-matrix electron-impact excitation data for the C-like iso-electronic sequence
    Rename them adas_ion.dat where ion runs on n1, o2, f3, ne4, na5, mg6, al7, and si8
    """
    url_base = "https://open.adas.ac.uk/download/adf04/copaw][c/clike_jm19]["
    url_file = ion + '.dat'
    url = url_base + url_file
    print('Download of', url, 'started')
    r = requests.get(url)
    with open('adas_' + url_file, 'wb') as f:
        f.write(r.content)
    print('Saving', 'adas_' + url_file, 'completed')
        
def make_pyneb(adas_filename, out_filename, ion_str, elem, ion, gsconf, NLevels=6):
    """
    Transform the ADAS adf04 data file into a pyneb file
    """
    use_data = False
    temp_str = None
    temp  = []
    I = []
    J = []
    Oms = []

    with open(adas_filename, 'r') as f:
        for l in f:
            if l == '   -1\n': # the data starts after this
                use_data = True
                continue
            if l == '  -1\n': # no more data after this
                break
            if use_data: # we read the data here
                if temp_str is None: # first data is temperature
                    temp_str = l.split()[2:]
                    for tt in temp_str:
                        temp.append(tt[0:-3] + 'e' + tt[-3:]) #add exponent (missing in original data)
                    continue
                dsplit = l.split() # now run on collision data
                I.append(int(dsplit[0])) # first field
                J.append(int(dsplit[1])) # secind field
                dsplit = dsplit[2:] # rest of the fields
                if len(dsplit[-1]) > 8: # the latest field can be 2 numbers together (because the 2nd one can be negative)
                    dd = dsplit[-1]
                    dsplit[-1] = dd[:7]
                    dsplit.append(dd[7:])
                dsplit2 = []
                for dd in dsplit:
                    dsplit2.append(dd[0:-3] + 'e' + dd[-3:]) # add exponent (missing in original data)
                Oms.append(dsplit2[1:-1])        
    temp = np.array(temp, dtype=float)
    Oms = pd.DataFrame(Oms, dtype=float)
    Oms.insert(loc=0, column='I', value=I)   # I and J are switched in PyNeb format
    Oms.insert(loc=0, column='J', value=J)   
    mask = Oms.I <= NLevels

    with open(out_filename, 'w') as f: # write the PyNeb data file
        f.write(' 0  0 ' + ' '.join(['{:.2e}'.format(float(tt)) for tt in temp]) + '\n')
        f.write(Oms[mask].to_string(index=False, header=False, float_format='{:.2e}'.format))
        f.write("""
*** {} collision strengths 
*** SOURCE1  J. Mao, N. R. Badnell, and G. Del Zanna 2020, A&A, 634, A7
*** NOTE1    Collision strengths
*** ATOM {}
*** SPECTRUM {}
*** N_LEVELS {}
*** GSCONFIG {} 
*** T_UNIT K
""".format(ion_str, elem, ion, NLevels, gsconf))
    print(out_filename, 'created from', adas_filename)

def download_all():
    """
    Download 8 files from ADAS
    """
    for ion in ('n1', 'o2', 'f3', 'ne4', 'na5', 'mg6', 'al7','si8'):
        download_file(ion)
        
def make_all(NLevels=9):
    """
    Write the 8 files in the PyNeb format
    """
    
    make_pyneb('adas_n1.dat','n_ii_coll_MBZ20.dat', 'N II', 'nitrogen', '2', 'p2', NLevels)
    make_pyneb('adas_o2.dat','o_iii_coll_MBZ20.dat', 'O III', 'oxygen', '3', 'p2', NLevels)
    make_pyneb('adas_f3.dat','f_iv_coll_MBZ20.dat', 'F IV', 'fluorine', '4', 'p2', NLevels)
    make_pyneb('adas_ne4.dat','ne_v_coll_MBZ20.dat', 'Ne V', 'neon', '5', 'p2', NLevels)
    make_pyneb('adas_na5.dat','na_vi_coll_MBZ20.dat', 'Na VI', 'sodium', '6', 'p2', NLevels)
    make_pyneb('adas_mg6.dat','mg_vii_coll_MBZ20.dat', 'Mg VII', 'magnesium', '7', 'p2', NLevels)
    make_pyneb('adas_al7.dat','al_viii_coll_MBZ20.dat', 'Al VIII', 'aluminium', '8', 'p2', NLevels)
    make_pyneb('adas_si8.dat','si_ix_coll_MBZ20.dat', 'Si IX', 'silicon', '9', 'p2', NLevels)



import numpy as np
import os
import sys
import traceback
import pyneb as pn

#from scipy.linalg import solve as solve_sc
from numpy.linalg import solve as solve_np


try:
    import cvxopt
    cvxopt_ok = True
    
except:
    cvxopt_ok = False

"""
from scipy.sparse.linalg import spsolve as solve_sp

try:
    from numba import double
    from numba.decorators import jit, autojit
except:
    pass


try:
    import rpy2.robjects.numpy2ri
    from rpy2.robjects.packages import importr
    rpy2.robjects.numpy2ri.activate()
    base     = importr('base')
    rpy_ok = True
except:
    rpy_ok = False
"""
    


# It seems we do not need anymore bs to deal with strings between py2 and py3
if sys.version_info.major < 3:
    def bs(x):
        return x
else:
    def bs(x):
        if isinstance(x, bytes):
            return x.decode(encoding='UTF-8')
        elif isinstance(x, str):
            return x.encode(encoding='UTF-8')

ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir))
      
def execution_path(filename):
    return os.path.join(os.path.dirname(sys._getframe(1).f_code.co_filename), filename)

def _returnNone(*argv, **kwargs):
    return None

def int_to_roman(input_):
    """
    Convert an integer to Roman numerals.
    
    Examples:
    >>> int_to_roman(0)
    Traceback (most recent call last):
    ValueError: Argument must be between 1 and 3999
    
    >>> int_to_roman(-1)
    Traceback (most recent call last):
    ValueError: Argument must be between 1 and 3999
    
    >>> int_to_roman(1.5)
    Traceback (most recent call last):
    TypeError: expected integer, got <type 'float'>
    
    >>> print int_to_roman(2000)
    MM

    >>> print int_to_roman(1999)
    MCMXCIX

    """
    if type(input_) != type(1):
        #raise TypeError, "expected integer, got %s" % type(input_)
        return None
    if not 0 < input_ < 4000:
        #raise ValueError, "Argument must be between 1 and 3999"
        return None
    ints = (1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1)
    nums = ('M', 'CM', 'D', 'CD', 'C', 'XC', 'L', 'XL', 'X', 'IX', 'V', 'IV', 'I')
    result = ""
    for i in range(len(ints)):
        count = int(input_ / ints[i])
        result += nums[i] * count
        input_ -= ints[i] * count
    return result


def roman_to_int(input_):
    """
    Convert a roman numeral to an integer.
    
    >>> r = range(1, 4000)
    >>> nums = [int_to_roman(i) for i in r]
    >>> ints = [roman_to_int(n) for n in nums]
    >>> print r == ints
    1
    
    >>> roman_to_int('VVVIV')
    Traceback (most recent call last):
     ...
    ValueError: input is not a valid roman numeral: VVVIV

    >>> roman_to_int(1)
    Traceback (most recent call last):
     ...
    TypeError: expected string, got <type 'int'>

    >>> roman_to_int('a')
    Traceback (most recent call last):
     ...
    ValueError: input is not a valid roman numeral: A

    >>> roman_to_int('IL')
    Traceback (most recent call last):
     ...
    ValueError: input is not a valid roman numeral: IL

    """   
    if type(input_) != type(""):
        #raise TypeError, "expected string, got %s" % type(input_)
        return None
    input_ = input_.upper()
    nums = ['M', 'D', 'C', 'L', 'X', 'V', 'I']
    ints = [1000, 500, 100, 50, 10, 5, 1]
    places = []
    for c in input_:
        if not c in nums:
            #raise ValueError, "input is not a valid roman numeral: %s" % input_
            return None
    for i in range(len(input_)):
        c = input_[i]
        value = ints[nums.index(c)]
        # If the next place holds a larger number, this value is negative.
        try:
            nextvalue = ints[nums.index(input_[i + 1])]
            if nextvalue > value:
                value *= -1
        except IndexError:
            # there is no next place.
            pass
        places.append(value)
    sum_ = 0
    for n in places: sum_ += n
    # Easiest test for validity...
    if int_to_roman(sum_) == input_:
        return sum_
    else:
        #raise ValueError, 'input is not a valid roman numeral: %s' % input_
        return None
      
      
def parseAtom(atom):
    '''
    Parses an atom label into the element and spectrum parts
    '''

    iso = ''
    elem = ''
    spec = ''
    cont = True
    firstdigit = True
    for l in atom:
        if l.isalpha() and cont:
            elem += l
            firstdigit = False
        elif l.isdigit():
            if firstdigit:
                iso += l
            else:
                spec += l
                cont = False
    return iso+str.capitalize(elem), spec

def parseAtom2(atom):
    '''
    Parses an atom label into the element and spectrum parts
    '''

    iso = ''
    elem = ''
    spec = ''
    cont = True
    firstdigit = True
    for l in atom:
        if l.isalpha() and cont:
            elem += l
            firstdigit = False
        elif l.isdigit():
            if firstdigit:
                iso += l
            else:
                spec += l
                cont = False
    if atom[-1] == 'r':
        rec = 'r'
    else:
        rec = ''
    return iso+str.capitalize(elem), spec, rec




def strExtract(text, par1=None, par2=None): 
    """
    Extract a substring from text (first parameter)
     
    If par1 is a string, the extraction starts after par1,
    else if it is an integer, it starts at position par1.
    If par 2 is a string, extraction stops at par2, 
    else if par2 is an integer, extraction stops after par2 characters.
    
    Examples: 
    
    strExtract('test123','e','1')
    strExtract('test123','st',4)
    
    """
    if np.size(text) == 1:
        if type(par1) is int:
            str1 = text[par1::]
        elif type(par1) is str:
            str1 = text.split(par1)[-1]
        else:
            str1 = text
    
        if type(par2) is int:
            str2 = str1[0:par2]
        elif type(par2) is str:
            str2 = str1.split(par2)[0]
        else:
            str2 = str1
        return str2
    else:
        res = []
        for subtext in text:
            res1 = strExtract(subtext, par1=par1, par2=par2)
            if res1 != '':
                res.append(res1)
        return res


def formatExceptionInfo(maxTBlevel=5):
    cla, exc, trbk = sys.exc_info()
    excName = cla.__name__
    try:
        excArgs = exc.__dict__["args"]
    except KeyError:
        excArgs = "<no args>"
    excTb = traceback.format_tb(trbk, maxTBlevel)
    return (excName, excArgs, excTb)

def multi_split(s, seps):
    res = [s]
    for sep in seps:
        s, res = res, []
        for seq in s:
            res += seq.split(sep)
    return res


def cleanPypicFiles(files=None, all_=False, dir_=None):
    """
    Method to clean the directory containing the pypics files.

    Parameters:
        - files    list of files to be removed
        - all_      Boolean, is set to True, all the files are deleted from the directory
        - dir_      directory from where the files are removed. 
                If None (default), pn.config.pypic_path is used.

    """
    if dir_ is None:
        dir_ = pn.config.pypic_path
    if all_:
        files = os.listdir(dir_)
        if np.ndim(files) == 0:
            files = [files]
    if files is None:
        return
    if type(files) == type(''):
        files = [files]
    for file_ in files:
        file_path = os.path.join(dir_, file_)
        try:
            if os.path.isfile(file_path):
                os.remove(file_path)
                pn.log_.message('Deleting {0}'.format(file_path), calling='cleanPypicFiles')
        except:
            pn.log_.warn('Unable to remove {0}'.format(file_path), calling='cleanPypicFiles')


def getPypicFiles(dir_=None):
    """
    Return the list of files in the directory.

    Parameters:
        - dir_     directory from where the files are removed. 
                   If None (default), pn.pypic_path is used.

    """
    if dir_ is None:
        dir_ = pn.config.pypic_path
    files = os.listdir(dir_)
    return files


def revert_seterr(oldsettings):
    """
    This function revert the options of seterr to a value saved in oldsettings.
    
    Usage:
        oldsettings = np.seterr(all='ignore')
        to_return = (result - int_ratio) / int_ratio # this will not issue Warning messages
        revert_seterr(oldsettings)

    Parameters:
        oldsettings  result of np.seterr(all='ignore')

    """
    np.seterr(over=oldsettings['over'])
    np.seterr(divide=oldsettings['divide'])
    np.seterr(invalid=oldsettings['invalid'])
    np.seterr(under=oldsettings['under'])

def quiet_divide(a, b):
    """
    This function returns the division of a by b, without any waring in case of b beeing 0.
    """
    oldsettings = np.seterr(all='ignore')
    to_return = a / b # this will not issue Warning messages
    revert_seterr(oldsettings)
    return to_return

def quiet_log10(a):
    """
    This function returns the log10 of a, without any waring in case of b beeing 0.
    """
    oldsettings = np.seterr(all='ignore')
    to_return = np.log10(a) # this will not issue Warning messages
    revert_seterr(oldsettings)
    return to_return
     
def get_reduced(N_rand, a, value_method = 'original', error_method='std'):
    """
    This function returns a tuple of value and error corresponding to an array of values 
        obtained from a MonteCarlo computation
    It takes as argument an array that contains the original value, 
        followed by N_rand MonteCarlo values.
    The relevant value is computed by returning the original value (default), the mean or the median, 
        depending on "value_method"
    The errors are computed by 68% quantiles or standart deviation (Default), 
        depending on "error_method"
    
    """

    if error_method == 'quants' and not pn.config.INSTALLED['scipy']:
        pn.log_.error('Scipy not installed, use_quants not available', calling = 'get_reduced')
    else:
        from scipy.stats.mstats import mquantiles
    
    
    if value_method == 'original':
        value = a[0]
    elif value_method == 'mean':
        value = a.mean()
    elif value_method == 'median':
        value = np.median(a)
    
    if error_method == 'quants':
        quants = mquantiles(a, [0.16, 0.84])
        error = (quants[1]-quants[0])/2.
    elif error_method == 'uquants':
        quants = mquantiles(a, [0.16, 0.84])
        error = quants[1] - value
    elif error_method == 'lquants':
        quants = mquantiles(a, [0.16, 0.84])
        error = value - quants[0]
        # error = (quants[1]-quants[0])/2)
    elif error_method == 'upper':
        mask = (a - value) >= 0
        error =  ((((a[mask] - value)**2).sum())/np.float(mask.sum()))**0.5
    elif error_method == 'lower':
        mask = (a - value) <= 0
        error = -((((a[mask] - value)**2).sum())/np.float(mask.sum()))**0.5
    elif error_method == 'std':
        error = a.std()
    else:
        pn.log_.error('Unknown error method {}'.format(error_method))
        
    return (value, error)
    
def get_reduced_dic(N_rand, n_obs_ori, dic, value_method = 'original', 
                    error_method='std', abund12 = False):
    """
    This function returns a dictionary of values (ionic or atomic abundances, temp...) and errors.
    It takes as argument a dictionary that is supposed to contains n_obs_ori values from
        the original observations, followed (optionaly) by N_rand*n_obs_ori MonteCarlo values.
    The new values are computed by returning the original value (default), the mean or the median, 
        depending on "value_method"
    The errors are computed by 68% quantiles or standart deviation (Default), depending on on "error_method"
    
    It also transforms the abundances by number into 12+log10(abundances by number)
    """
    res = {}
    for key in dic.keys():
        value = []
        error = []
        for i in np.arange(n_obs_ori):
            if abund12:
                values = 12 + np.log10(dic[key][n_obs_ori+i*N_rand:n_obs_ori+(i+1)*N_rand])
                values = np.insert(values, 0, 12 + np.log10(dic[key][i]))
            else:
                values = dic[key][n_obs_ori+i*N_rand:n_obs_ori+(i+1)*N_rand]
                values = np.insert(values, 0, dic[key][i])
            tt = np.isfinite(values)
            if tt.sum() == 0:
                value.append(np.NAN)
                error.append(0.0)
            else:
                v, e = get_reduced(N_rand, values[tt], value_method = value_method, 
                                   error_method=error_method)
                value.append(v)
                error.append(e)
        res[key] = np.array(value)
        res['{0}_e'.format(key)] = np.array(error)
    return res

def addRand(N, list_, list_errors=None, lowlim=None):
    """
    This function adds MonteCarlo random-gauss values to a list.
    
    Parameters:
        N: number of MonteCarlo values to be added
        list_: list of input values
        list_errors: list of errors associated to the values in list_. The errors are absolutes, the 
            result will be a gaussian distribution of sigma=list_errors.
            If None, the functions adds N time the same values to the list
        lowlim: if not None (default), any random-gauss number lower then the lowlim value is set to lowlim.
        
    Usage:
        print addRand(3, [1,2,3]) # no errors: no random values, only replicates the values
            [1, 2, 3, 1, 1, 1, 2, 2, 2, 3, 3, 3]
        print addRand(3, [1,20,300], [1, 1, 0.1])
            [1, 20, 300, 
            1.550094377016822, 1.868307356917796, 1.0242090163674675, 
            19.782857703031276, 19.049474190752157, 21.58680361755194, 
            299.99810384362934, 300.00753905080154, 299.94052054137694]
    """
    
    # Check that the 2nd argument is a list
    if type(list_) != type([]):
        pn.log_.error('The second argument must be a list', calling = 'addRand')
    new_list = list_[:]
    # Computes the new values
    if N != 0:
        # initialize the new list with the values of list_
        to_extend = []
        for i in range(len(list_)):
            if list_errors is None:
                to_extend.extend([list_[i]] * N)
            else:
                to_extend.extend(np.random.standard_normal(N) * list_errors[i] + list_[i])
        # filter the values lower than lowlim
        if lowlim is not None:
            to_extend = [value if value > lowlim else lowlim for value in to_extend]
        new_list.extend(to_extend)
    return new_list

"""
def solve_r(a, b):
    if rpy_ok:
        return base.solve(a, b)
    else:
        return None
    
def solve_lapack(a, b):
    from numpy.linalg import lapack_lite

    n_eq = a.shape[0]
    n_rhs = b.shape[0]

    pivots = np.zeros(n_eq, np.intc)
    results = lapack_lite.dgesv(n_eq, n_rhs, a, n_eq, pivots, b, n_eq, 0)
    return results  
"""

if cvxopt_ok:
    def solve_cvxopt(a, b):
        A = cvxopt.matrix(a)
        B = cvxopt.matrix(b)
        return cvxopt.lapack.gesv(A, B)    
    
"""
# @profile
if cvxopt_ok:
    @profile
    def solve(a, b):
        return solve_cvxopt(a, b)
else:
    @profile
    def solve(a, b):
        return solve_np(a,b)
"""
#@profile
def solve(a, b):
    return solve_np(a,b)

